"""View representing the result of a Query."""

from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from html import escape as htmlescape
from typing import TYPE_CHECKING, Any, TypeVar, cast, overload

import numpy as np
from pydantic import BaseModel
from tabulate import tabulate

from ultimate_notion import props, schema
from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.file import ExternalFile, NotionFile
from ultimate_notion.obj_api.enums import FormulaType, RollupType
from ultimate_notion.page import Page
from ultimate_notion.props import Option
from ultimate_notion.rich_text import html_img
from ultimate_notion.schema import Property
from ultimate_notion.user import User
from ultimate_notion.utils import SList, deepcopy_with_sharing, find_index, find_indices, is_notebook, rec_apply

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    from numpy.typing import NDArray

    from ultimate_notion.database import Database
    from ultimate_notion.props import PropertyValue
    from ultimate_notion.query import Query

T = TypeVar('T')


class View(Sequence[Page]):
    def __init__(self, database: Database, pages: Sequence[Page], query: Query):
        self.database = database
        self._query = query
        self._title_col = database.schema.get_title_prop().name
        self._columns = self._get_columns(self._title_col)
        self._pages: NDArray[Any] = np.array(pages)
        self._row_indices: NDArray[Any]
        self._col_indices: NDArray[Any]
        self.default_limit = 10

        self.reset()

    def reset(self) -> View:
        """Reset the view, i.e. remove filtering, index and sorting."""
        self._icon_name: str | None = None
        self._id_name: str | None = None
        self._index_name: str | None = None
        self._row_indices = np.arange(len(self._pages))
        self._col_indices = np.arange(len(self._columns))
        return self

    def clone(self) -> View:
        """Clone the current view."""
        return deepcopy_with_sharing(self, shared_attributes=['database', '_pages', '_query'])

    def _get_columns(self, title_col: str) -> NDArray[Any]:
        """Make sure title column is the first columns."""
        cols = list(self.database.schema.to_dict().keys())
        cols.insert(0, cols.pop(cols.index(title_col)))
        return np.array(cols)

    @property
    def columns(self) -> list[str]:
        """Columns/properties of the database view aligned with the elements of a row."""
        cols = list(self._columns[self._col_indices])
        if self.has_icon:
            cols.insert(0, self._icon_name)
        if self.has_id:
            cols.insert(0, self._id_name)
        if self.has_index:
            cols.insert(0, self._index_name)
        return cols

    @overload
    def __getitem__(self, idx: int, /) -> Page: ...

    @overload
    def __getitem__(self, idx: slice, /) -> Sequence[Page]: ...

    def __getitem__(self, idx: int | slice, /) -> Page | Sequence[Page]:
        pages = self._pages[self._row_indices[idx]]
        if isinstance(idx, slice):
            return tuple(pages)
        else:
            return cast(Page, pages)

    def get_page(self, idx: int, /) -> Page:
        """Retrieve a page by index of the view."""
        return cast(Page, self._pages[self._row_indices[idx]])

    def search_page(self, name: str) -> SList[Page]:
        """Retrieve a page from this view by name"""
        pages = [page for page in self.to_pages() if page.title == name]
        return SList(pages)

    def get_row(self, idx: int, /) -> tuple[Any, ...]:
        """Retrieve a row, i.e. all properties of a page defined by an index."""
        page = self.get_page(idx)
        row: list[Any] = []
        for col in self.columns:
            if col == self._title_col:
                row.append(page.title)
            elif col == self._id_name:
                row.append(page.id)
            elif col == self._index_name:
                row.append(idx)
            elif col == self._icon_name:
                row.append(page.icon)
            else:
                row.append(page.props[col])
        return tuple(row)

    def to_pages(self) -> list[Page]:
        """Convert the view to a simple list of pages."""
        return [self.get_page(idx) for idx in range(len(self))]

    def to_rows(self) -> list[tuple[Any, ...]]:
        return [self.get_row(idx) for idx in range(len(self))]

    def to_pydantic(self) -> list[BaseModel]:
        """Convert the view to a list of Pydantic models."""
        model = self.database.schema.to_pydantic_model(with_ro_props=True)
        is_prop_ro = {prop.name: prop.readonly for prop in self.database.schema.get_props()}

        def get_prop(page: Page, prop: str) -> PropertyValue | Any:
            prop_value = page.props._get_property(prop)
            return prop_value if is_prop_ro[prop] else prop_value.value

        models: list[BaseModel] = []
        for page in self.to_pages():
            models.append(model(**{col: get_prop(page, col) for col in self.columns}))

        return models

    def to_pandas(self) -> pd.DataFrame:
        """Convert the view to a Pandas dataframe."""
        import pandas as pd  # noqa: PLC0415

        # remove index as pandas uses its own
        view = self.without_index() if self.has_index else self
        data = rec_apply(cmplx_to_str, self.to_rows())

        return pd.DataFrame(data, columns=view.columns)

    def _probe_col_type(self, col: str) -> PropertyValue | None:
        """Probe the type of a column by checking the values in the column."""
        for page in self.to_pages():
            match prop := page.props._get_property(col):
                case props.Relations():
                    # we can only determine the type of a relation if it has at least one value
                    if prop.value:
                        return prop
                case _:
                    return prop
        return None

    def _to_polars_schema(self) -> dict[str, pl.DataType]:
        """Create a Polars schema for the view."""
        import polars as pl  # noqa: PLC0415

        db_schema = self.database.schema.to_dict()
        pl_schema: dict[str, pl.DataType] = {}
        for col in self.columns:
            if col in db_schema:
                prop_type: Property | PropertyValue = db_schema[col]
                if isinstance(prop_type, schema.Rollup | schema.Relation | schema.Formula):
                    prop_type = probed_type if (probed_type := self._probe_col_type(col)) is not None else prop_type
                pl_schema[col] = prop_type_to_polars(prop_type)
            elif col == self._icon_name:
                pl_schema[col] = pl.String()
            elif col == self._index_name:
                pl_schema[col] = pl.Int64()
            elif col == self._id_name:
                pl_schema[col] = pl.String()
            else:
                msg = f'Column `{col}` not found in database schema.'
                raise RuntimeError(msg)
        return pl_schema

    def to_polars(self) -> pl.DataFrame:
        """Convert the view to a Polars dataframe."""
        import polars as pl  # noqa: PLC0415

        data = rec_apply(cmplx_to_str, self.to_rows())
        schema = self._to_polars_schema()
        return pl.DataFrame(data=data, schema=schema, orient='row')

    def _html_for_icon(self, rows: list[Any], cols: list[str]) -> list[Any]:
        # escape everything as we ask tabulate not to do it
        rows = [[htmlescape(elem) if isinstance(elem, str) else elem for elem in row] for row in rows]
        if (title_idx := find_index(self._icon_name, cols)) is None:
            return rows
        for idx, row in enumerate(rows):
            page = self.get_page(idx)
            icon = html_img(page.icon.url, size=1.2) if isinstance(page.icon, NotionFile | ExternalFile) else page.icon
            row[title_idx] = icon
        return rows

    def as_table(self, tablefmt: str | None = None) -> str:
        """Return the view in a given string table format.

        Some table formats:

        - plain: no pseudographics
        - simple: Pandoc's simple table, i.e. only dashes to separate header from content
        - github: GitHub flavored Markdown
        - simple_grid: uses dashes & pipes to separate cells
        - html: standard html markup

        Find more table formats under: https://github.com/astanin/python-tabulate#table-format
        """
        rows = self.to_rows()
        cols = self.columns

        if tablefmt is None:
            tablefmt = 'html' if is_notebook() else 'simple'

        if tablefmt == 'html':
            if self.has_icon:
                rows = self._html_for_icon(rows, cols)
                html_str = str(tabulate(rows, headers=cols, tablefmt='unsafehtml'))  # str() as tabulate wraps the str
            else:
                html_str = str(tabulate(rows, headers=cols, tablefmt='html'))  # str() as tabulate wraps the str
            return html_str
        else:
            return tabulate(rows, headers=cols, tablefmt=tablefmt)

    def show(self, *, simple: bool | None = None) -> None:
        """Show the database as human-readable table."""
        if simple:
            tablefmt = 'simple'
        elif simple is None:
            tablefmt = 'html' if is_notebook() else 'simple'
        else:
            tablefmt = 'html'

        table_str = self.as_table(tablefmt=tablefmt)

        if is_notebook() and (tablefmt == 'html'):
            from IPython.display import display_html  # noqa: PLC0415

            display_html(table_str)
        else:
            print(table_str)  # noqa: T201

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this view."""
        if len(self) > self.default_limit:
            html_str = self.limit(self.default_limit).as_table(tablefmt='html')
            html_str = f'<div style="display: inline-block;text-align: center;">{html_str}&#8942;</div></div>\n'
            return html_str
        else:
            return self.as_table(tablefmt='html')

    def __iter__(self) -> Iterator[Page]:
        """Iterate over the pages in the view."""
        return iter(self.to_pages())

    def __repr__(self) -> str:
        return get_repr(self, desc=self.database.title)

    def __str__(self) -> str:
        return self.as_table()

    def __len__(self) -> int:
        return len(self._row_indices)

    @property
    def is_empty(self) -> bool:
        """Is this view empty?"""
        return len(self) == 0

    def __bool__(self) -> bool:
        """Overwrite default behaviour."""
        msg = 'Use .is_empty instead of bool(view) to check if a view is empty.'
        raise RuntimeError(msg)

    @property
    def has_index(self) -> bool:
        return self._index_name is not None

    def with_index(self, name: str = 'index') -> View:
        """Add an index column to the view."""
        if self.has_index and name == self._index_name:
            return self

        if name in self.columns:
            msg = f"index '{name}' is already a column name"
            raise RuntimeError(msg)

        view = self.clone()
        view._index_name = name
        return view

    def without_index(self) -> View:
        """Remove index column from the view."""
        if not self.has_index:
            return self

        view = self.clone()
        view._index_name = None
        return view

    @property
    def has_icon(self) -> bool:
        return self._icon_name is not None

    def with_icon(self, name: str = 'icon') -> View:
        """Show icons in HTML output."""
        if self.has_icon and name == self._icon_name:
            return self

        view = self.clone()
        view._icon_name = name
        return view

    def without_icon(self) -> View:
        """Don't show icons in HTML output."""
        if not self.has_icon:
            return self

        view = self.clone()
        view._icon_name = None
        return view

    @property
    def has_id(self) -> bool:
        return self._id_name is not None

    def with_id(self, name: str = 'id') -> View:
        """Add an id column to the view."""
        if self.has_id and name == self._id_name:
            return self

        view = self.clone()
        view._id_name = name
        return view

    def without_id(self) -> View:
        """Remove id column from the view."""
        if not self.has_id:
            return self

        view = self.clone()
        view._id_name = None
        return view

    def head(self, num: int) -> View:
        """Keep only the first `num` elements in view."""
        view = self.clone()
        view._row_indices = view._row_indices[:num]
        return view

    def limit(self, num: int) -> View:
        """Alias for `head`"""
        return self.head(num)

    def tail(self, num: int) -> View:
        """Keep only the last `num` elements in view."""
        view = self.clone()
        view._row_indices = view._row_indices[-num:]
        return view

    def select(self, *cols: str) -> View:
        """Select columns for the view"""
        curr_cols = self._columns  # we only consider non-meta columns, e.g. no index, etc.
        if not_included := set(cols) - set(curr_cols):
            msg = f'Some columns, i.e. {", ".join(not_included)}, are not in view'
            raise RuntimeError(msg)

        view = self.clone()
        select_col_indices = find_indices(list(cols), curr_cols)
        view._col_indices = view._col_indices[select_col_indices]
        return view

    def apply(self, func: Callable[[Page], T]) -> list[T]:
        """Apply function to all pages in view.

        Args:
            func: function taking a Page as input
        """
        return [func(page) for page in self.to_pages()]

    def reverse(self) -> View:
        """Reverse the order of the rows."""
        view = self.clone()
        view._row_indices = view._row_indices[::-1]
        return view

    def sort(self) -> View:
        """Sort the view with respect to some columns."""
        # ToDo: Implement me
        raise NotImplementedError

    def filter(self) -> View:
        """Filter the view."""
        # ToDo: Implement me
        raise NotImplementedError

    def reload(self) -> View:
        """Reload all pages by re-executing the query that generated the view."""
        view = self.clone()
        view._pages = np.array(self._query.execute().to_pages())
        return view


def cmplx_to_str(obj: Wrapper) -> Wrapper | str:
    """Convert complex objects to a string representation."""
    match obj:
        case NotionFile() | ExternalFile():
            return obj.url
        case Option():
            return obj.name
        case User():
            return str(obj)
        case _:
            return obj


def prop_type_to_polars(prop_valtype: Property | PropertyValue) -> pl.DataType:
    """Convert a Notion property type to a Polars data type."""
    import polars as pl  # noqa: PLC0415

    match prop_valtype:
        case schema.Title() | schema.Text() | props.Title() | props.Text():
            return pl.String()
        case schema.ID() | props.ID():
            return pl.String()
        case schema.Verification() | props.Verification():
            return pl.Object()
        case schema.Button() | props.Button():
            return pl.Object()
        case schema.Number() | props.Number():
            return pl.Float64()
        case schema.Select() as prop:
            return pl.Enum([opt.name for opt in prop.options])
        case props.Select():
            return pl.Categorical()
        case schema.MultiSelect() as prop:
            return pl.List(pl.Enum([opt.name for opt in prop.options]))
        case schema.MultiSelect():
            return pl.List(pl.Categorical)
        case schema.Checkbox() | props.Checkbox():
            return pl.Boolean()
        case schema.Date() | props.Date():
            return pl.Datetime(time_unit='ms')
        case schema.Files() | schema.Person() | props.Files() | props.Person():
            # Fix 2025-06-06: `not yet implemented: Nested object types` pl.List(pl.Object()) not working
            return pl.List(pl.String())
        case schema.CreatedTime() | schema.LastEditedTime() | props.CreatedTime() | props.LastEditedTime():
            return pl.Datetime(time_unit='ms')
        case schema.CreatedBy() | schema.LastEditedBy() | props.CreatedBy() | props.LastEditedBy():
            return pl.Object()
        case schema.URL() | schema.Email() | schema.Phone() | props.URL() | props.Email() | props.Phone():
            return pl.String()
        case schema.Status() as prop:
            return pl.Enum([opt.name for opt in prop.options])
        case props.Status():
            return pl.Categorical()
        case props.Formula() as prop:
            match prop.value_type:
                case FormulaType.NUMBER:
                    return pl.Float64()
                case FormulaType.STRING:
                    return pl.String()
                case FormulaType.BOOLEAN:
                    return pl.Boolean()
                case FormulaType.DATE:
                    return pl.Datetime(time_unit='ms')
                case _:
                    msg = f'Unsupported formula type: {prop.value_type}'
                    raise ValueError(msg)
        case props.Rollup() as prop:
            match prop.value_type:
                case RollupType.NUMBER:
                    return pl.Float64()
                case RollupType.DATE:
                    return pl.Datetime(time_unit='ms')
                case RollupType.ARRAY:
                    # Fix 2025-06-06: `not yet implemented: Nested object types` pl.List(pl.Object()) not working
                    return pl.List(pl.String())
                case _:
                    msg = f'Unsupported rollup type: {prop.value_type}'
                    raise ValueError(msg)

        case schema.Relation() | schema.Rollup() | schema.Formula():
            # fallback as we cannot convert due to a lack of actual type information in the schema types
            return pl.Object()
        case _:
            msg = f'Unsupported property type or value: {prop_valtype}'
            raise ValueError(msg)



"""All definitions and enumerations"""

from __future__ import annotations

from enum import Enum

# ToDo: Let Enums inherit form StrEnum instead of Enum & str, when we require Python >= 3.11


class SortDirection(str, Enum):
    """Sort direction options."""

    ASCENDING = 'ascending'
    DESCENDING = 'descending'


class TimestampKind(str, Enum):
    """Possible timestamp types."""

    CREATED_TIME = 'created_time'
    LAST_EDITED_TIME = 'last_edited_time'


class VState(str, Enum):
    """Verification states for pages in wiki databases."""

    VERIFIED = 'verified'
    UNVERIFIED = 'unverified'


class Color(str, Enum):
    """Basic colors

    DEFAULT is a light gray, which is the default color in the Notion App.
    """

    DEFAULT = 'default'
    GRAY = 'gray'
    BROWN = 'brown'
    ORANGE = 'orange'
    YELLOW = 'yellow'
    GREEN = 'green'
    BLUE = 'blue'
    PURPLE = 'purple'
    PINK = 'pink'
    RED = 'red'


class FormulaType(str, Enum):
    """Formula types for formulas.

    The type of the formula as well as the mapping to the Notion API keyword when filtering in a query.
    """

    NUMBER = 'number', 'number'
    STRING = 'string', 'string'
    BOOLEAN = 'boolean', 'checkbox'  # inconsistency in the Notion API
    DATE = 'date', 'date'

    formula_kwarg: str

    def __new__(cls, value: str, formula_kwarg: str = '') -> FormulaType:
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.formula_kwarg = formula_kwarg
        return obj


class RollupType(str, Enum):
    """Rollup types for rollups."""

    NUMBER = 'number'
    DATE = 'date'
    ARRAY = 'array'
    INCOMPLETE = 'incomplete'
    UNSUPPORTED = 'unsupported'


class ArrayQuantifier(str, Enum):
    """Array quantifiers for rollups."""

    ANY = 'any'
    EVERY = 'every'
    NONE = 'none'


class OptionGroupType(str, Enum):
    """Option group type of the Status property."""

    TO_DO = 'to_do'
    IN_PROGRESS = 'in_progress'
    COMPLETE = 'complete'


class BGColor(str, Enum):
    """Background colors for most textual blocks, e.g. paragraphs, callouts, etc."""

    DEFAULT = 'default'
    GRAY = 'gray_background'
    BROWN = 'brown_background'
    ORANGE = 'orange_background'
    YELLOW = 'yellow_background'
    GREEN = 'green_background'
    BLUE = 'blue_background'
    PURPLE = 'purple_background'
    PINK = 'pink_background'
    RED = 'red_background'


class FileUploadStatus(str, Enum):
    """Status of a file upload."""

    PENDING = 'pending'
    UPLOADED = 'uploaded'
    EXPIRED = 'expired'
    FAILED = 'failed'


class FileUploadMode(str, Enum):
    """File upload modes."""

    SINGLE_PART = 'single_part'
    MULTI_PART = 'multi_part'
    EXTERNAL_URL = 'external_url'


class CodeLang(str, Enum):
    """Coding languages for code blocks."""

    ABAP = 'abap'
    ARDUINO = 'arduino'
    BASH = 'bash'
    BASIC = 'basic'
    C = 'c'
    CLOJURE = 'clojure'
    COFFEESCRIPT = 'coffeescript'
    CPP = 'c++'
    CSHARP = 'c#'
    CSS = 'css'
    DART = 'dart'
    DIFF = 'diff'
    DOCKER = 'docker'
    ELIXIR = 'elixir'
    ELM = 'elm'
    ERLANG = 'erlang'
    FLOW = 'flow'
    FORTRAN = 'fortran'
    FSHARP = 'f#'
    GHERKIN = 'gherkin'
    GLSL = 'glsl'
    GO = 'go'
    GRAPHQL = 'graphql'
    GROOVY = 'groovy'
    HASKELL = 'haskell'
    HTML = 'html'
    JAVA = 'java'
    JAVASCRIPT = 'javascript'
    JSON = 'json'
    JULIA = 'julia'
    KOTLIN = 'kotlin'
    LATEX = 'latex'
    LESS = 'less'
    LISP = 'lisp'
    LIVESCRIPT = 'livescript'
    LUA = 'lua'
    MAKEFILE = 'makefile'
    MARKDOWN = 'markdown'
    MARKUP = 'markup'
    MATLAB = 'matlab'
    MERMAID = 'mermaid'
    NIX = 'nix'
    OBJECTIVE_C = 'objective-c'
    OCAML = 'ocaml'
    PASCAL = 'pascal'
    PERL = 'perl'
    PHP = 'php'
    PLAIN_TEXT = 'plain text'
    POWERSHELL = 'powershell'
    PROLOG = 'prolog'
    PROTOBUF = 'protobuf'
    PYTHON = 'python'
    R = 'r'
    REASON = 'reason'
    RUBY = 'ruby'
    RUST = 'rust'
    SASS = 'sass'
    SCALA = 'scala'
    SCHEME = 'scheme'
    SCSS = 'scss'
    SHELL = 'shell'
    SQL = 'sql'
    SWIFT = 'swift'
    TOML = 'toml'
    TYPESCRIPT = 'typescript'
    VB_NET = 'vb.net'
    VERILOG = 'verilog'
    VHDL = 'vhdl'
    VISUAL_BASIC = 'visual basic'
    WEBASSEMBLY = 'webassembly'
    XML = 'xml'
    YAML = 'yaml'
    MISC = 'java/c/c++/c#'


class AggFunc(str, Enum):
    """Aggregation functions for formulas.

    The categories naming corresonds mostly to the naming in the Notion App but
    also eliminates some inconsistencies. The first part of the tuple is the
    string used in the Notion API, the second part is an alias corresoinding to
    the defined category num.
    """

    SHOW_ORIGINAL = 'show_original', 'show_original'
    SHOW_UNIQUE = 'show_unique', 'show_unique'

    # Count
    COUNT_ALL = 'count', 'count_all'
    COUNT_VALUES = 'count_values', 'count_values'
    COUNT_UNIQUE_VALUES = 'unique', 'count_unique_values'
    COUNT_EMPTY = 'empty', 'count_empty'
    COUNT_NOT_EMPTY = 'not_empty', 'count_not_empty'
    COUNT_CHECKED = 'checked', 'count_checked'  # just called 'checked' in the Notion App
    COUNT_UNCHECKED = 'unchecked', 'count_unchecked'  # just called 'unchecked' in the Notion App
    COUNT_PER_GROUP = 'count_per_group', 'count_per_group'  # only for type 'status'

    # Percent
    PERCENT_EMPTY = 'percent_empty', 'percent_empty'
    PERCENT_NOT_EMPTY = 'percent_not_empty', 'percent_not_empty'
    PERCENT_CHECKED = 'percent_checked', 'percent_checked'
    PERCENT_PER_GROUP = 'percent_per_group', 'percent_per_group'  # only for type 'status'

    # Date
    DATE_RANGE = 'date_range', 'date_range'
    EARLIEST_DATE = 'earliest_date', 'earliest_date'
    LATEST_DATE = 'latest_date', 'latest_date'

    # More options, e.g. for numbers
    SUM = 'sum', 'sum'
    AVERAGE = 'average', 'average'
    MEDIAN = 'median', 'median'
    MIN = 'min', 'min'
    MAX = 'max', 'max'
    RANGE = 'range', 'range'

    alias: str

    def __new__(cls, value: str, alias: str = '') -> AggFunc:
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.alias = alias
        return obj

    @classmethod
    def from_alias(cls, alias: str) -> AggFunc:
        for item in cls:
            if item.alias == alias:
                return item
        msg = f'{alias} is not a valid {cls.__name__}'
        raise ValueError(msg)


class NumberFormat(str, Enum):
    """Number formats for numbers."""

    NUMBER = 'number'
    NUMBER_WITH_COMMAS = 'number_with_commas'
    PERCENT = 'percent'
    DOLLAR = 'dollar'
    CANADIAN_DOLLAR = 'canadian_dollar'
    AUSTRALIAN_DOLLAR = 'australian_dollar'
    EURO = 'euro'
    POUND = 'pound'
    YEN = 'yen'
    RUBLE = 'ruble'
    RUPEE = 'rupee'
    WON = 'won'
    YUAN = 'yuan'
    REAL = 'real'
    LIRA = 'lira'
    RUPIAH = 'rupiah'
    FRANC = 'franc'
    HONG_KONG_DOLLAR = 'hong_kong_dollar'
    NEW_ZEALAND_DOLLAR = 'new_zealand_dollar'
    KRONA = 'krona'
    NORWEGIAN_KRONE = 'norwegian_krone'
    MEXICAN_PESO = 'mexican_peso'
    RAND = 'rand'
    NEW_TAIWAN_DOLLAR = 'new_taiwan_dollar'
    DANISH_KRONE = 'danish_krone'
    ZLOTY = 'zloty'
    BAHT = 'baht'
    FORINT = 'forint'
    KORUNA = 'koruna'
    SHEKEL = 'shekel'
    CHILEAN_PESO = 'chilean_peso'
    PHILIPPINE_PESO = 'philippine_peso'
    DIRHAM = 'dirham'
    COLOMBIAN_PESO = 'colombian_peso'
    RIYAL = 'riyal'
    RINGGIT = 'ringgit'
    LEU = 'leu'
    ARGENTINE_PESO = 'argentine_peso'
    URUGUAYAN_PESO = 'uruguayan_peso'



"""Functionality to initialize a Notion Session."""

from __future__ import annotations

import io
import logging
import os
import time
from collections.abc import Sequence
from threading import RLock
from types import TracebackType
from typing import Any, BinaryIO, ClassVar, cast
from uuid import UUID

import httpx
import notion_client
from notion_client.errors import APIResponseError

from ultimate_notion.blocks import Block, DataObject, _append_block_chunks, _chunk_blocks_for_api
from ultimate_notion.config import Config, activate_debug_mode, get_or_create_cfg
from ultimate_notion.database import Database
from ultimate_notion.errors import SessionError, UnknownPageError, UnknownUserError
from ultimate_notion.file import MAX_FILE_SIZE, UploadedFile, get_file_size, get_mime_type
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import create_notion_client
from ultimate_notion.obj_api import query as obj_query
from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.obj_api.endpoints import NotionAPI
from ultimate_notion.obj_api.enums import FileUploadMode, FileUploadStatus
from ultimate_notion.obj_api.objects import UnknownUser as UnknownUserObj
from ultimate_notion.obj_api.objects import get_uuid
from ultimate_notion.page import Page
from ultimate_notion.props import Title
from ultimate_notion.rich_text import Text
from ultimate_notion.schema import DefaultSchema, Schema
from ultimate_notion.user import Bot, User
from ultimate_notion.utils import SList

_logger = logging.getLogger(__name__)


class Session:
    """A session for the Notion API.

    The session keeps tracks of all objects, e.g. pages, databases, etc.
    in an object store to avoid unnecessary calls to the API.
    """

    client: notion_client.Client
    api: NotionAPI
    _active_session: Session | None = None
    _lock = RLock()
    _own_bot_id: UUID | None = None
    cache: ClassVar[dict[UUID, DataObject | User]] = {}

    def __init__(self, cfg: Config | None = None, *, client: notion_client.Client | None = None, **kwargs: Any):
        """Initialize the `Session` object and the raw `api` endpoints.

        Args:
            cfg: configuration object
            **kwargs: Arguments for the [Notion SDK Client](https://ramnes.github.io/notion-sdk-py/reference/client/)
        """
        cfg = get_or_create_cfg() if cfg is None else cfg

        Session._initialize_once(self)

        if cfg.ultimate_notion.debug:
            activate_debug_mode()

        self.client = create_notion_client(cfg, **kwargs) if client is None else client
        self.api = NotionAPI(self.client)

    @classmethod
    def _initialize_once(cls, instance: Session) -> None:
        _logger.info('Initializing Notion session.')
        with Session._lock:
            if Session._active_session and Session._active_session is not instance:
                msg = 'Cannot initialize multiple Sessions at once'
                raise ValueError(msg)
            else:
                Session._active_session = instance

    @classmethod
    def get_active(cls) -> Session:
        """Return the current active session or None."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                msg = 'There is no activate Session'
                raise ValueError(msg)

    @classmethod
    def get_or_create(cls, *args: Any, **kwargs: Any) -> Session:
        """Return the current active session or create a new session."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                return cls(*args, **kwargs)

    def __enter__(self) -> Session:
        # Do nothing here as `__init__` created a client already and calling
        # `self.client.__enter__()` would initialize another client internally
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        self.close()

    def close(self) -> None:
        """Close the session and release resources."""
        _logger.info('Closing connection to Notion.')
        self.client.close()
        Session._active_session = None
        Session.cache.clear()
        Session._own_bot_id = None

    def is_closed(self) -> bool:
        """Determine if the session is closed or not."""
        try:
            self.raise_for_status()
        except RuntimeError:  # used by httpx client in case of closed connection
            return True
        else:
            return False

    def raise_for_status(self) -> None:
        """Confirm that the session is active and raise otherwise.

        Raises SessionError if there is a problem, otherwise returns None.
        """
        _logger.info('Checking connection to Notion.')
        try:
            self.whoami()
        except httpx.ConnectError as err:
            msg = 'Unable to connect to Notion'
            raise SessionError(msg) from err
        except APIResponseError as err:
            msg = 'Invalid API reponse'
            raise SessionError(msg) from err

    def get_block(self, block_ref: UUID | str, *, use_cache: bool = True) -> Block:
        """Retrieve a single block by an object reference."""
        block_uuid = get_uuid(block_ref)
        if use_cache and block_uuid in self.cache:
            _logger.info(f'Retrieving cached block with id `{block_uuid}`.')
            block = cast(Block, self.cache[block_uuid])
        else:
            _logger.info(f'Retrieving block with id `{block_uuid}`.')
            block = Block.wrap_obj_ref(self.api.blocks.retrieve(block_uuid))
            self.cache[block.id] = block
        _logger.info(f'Retrieved `{type(block)}` block.')
        return block

    def create_db(
        self, parent: Page, *, schema: type[Schema] | None = None, title: str | None = None, inline: bool = False
    ) -> Database:
        """Create a new database within a page.

        In case a title and a schema ware provided, title overrides the schema's `db_title` attribute if it exists.
        """
        # Implementation:
        # 1. create the database using a Notion API call and potential external forward relations
        # 2. initialize self-referencing forward relations
        # 3. create properties with self-referencing forward relations using an update call
        # 4. update the backward references, i.e. two-way relations, using an update call
        if title is not None:
            title = Text(title)
        elif schema is not None and schema._db_title is not None:
            title = schema._db_title
        else:
            title = None  # Anonymous database without a title.

        db_schema: type[Schema] = cast(type[Schema], DefaultSchema) if schema is None else schema
        _logger.info(f'Creating database `{title or "<NoTitle>"}` in page `{parent.title}` with schema:\n{db_schema}')

        title_obj = title.obj_ref if title is not None else None
        schema_dct = {prop.name: prop.obj_ref for prop in db_schema.get_props() if prop._is_init_ready}
        db_obj = self.api.databases.create(parent=parent.obj_ref, title=title_obj, schema=schema_dct, inline=inline)

        db: Database = Database.wrap_obj_ref(db_obj)

        if schema:
            if schema._db_desc:
                self.api.databases.update(db_obj, description=schema._db_desc.obj_ref)
            db._set_schema(schema, during_init=True)  # schema is thus bound to the database
            schema._init_self_refs()
            schema._init_self_ref_rollups()
            schema._update_bwd_rels()

        self.cache[db.id] = db
        return db

    def create_dbs(self, parents: Page | list[Page], schemas: list[type[Schema]]) -> list[Database]:
        """Create new databases in the right order in case there a relations between them."""
        # ToDo: Implement
        raise NotImplementedError

    def search_db(
        self, db_name: str | None = None, *, exact: bool = True, reverse: bool = False, deleted: bool = False
    ) -> SList[Database]:
        """Search a database by name or return all if `db_name` is None.

        Args:
            db_name: name/title of the database, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
            deleted: include deleted databases in search
        """
        _logger.info(f'Searching for database with name `{db_name}`.')
        query = cast(obj_query.SearchQueryBuilder[obj_blocks.Database], self.api.search(db_name).filter(db_only=True))
        if reverse:
            query.sort(ascending=True)
        dbs = [
            cast(Database, self.cache.setdefault(raise_unset(db.id), Database.wrap_obj_ref(db)))
            for db in query.execute()
        ]
        if exact and db_name is not None:
            dbs = [db for db in dbs if db.title == db_name]
        if not deleted:
            dbs = [db for db in dbs if not db.is_deleted]
        return SList(dbs)

    def get_db(self, db_ref: UUID | str, *, use_cache: bool = True) -> Database:
        """Retrieve Notion database by uuid"""
        db_uuid = get_uuid(db_ref)
        if use_cache and db_uuid in self.cache:
            _logger.info(f'Retrieving cached database with id `{db_uuid}`.')
            db = cast(Database, self.cache[db_uuid])
        else:
            _logger.info(f'Retrieving database with id `{db_uuid}`.')
            db = Database.wrap_obj_ref(self.api.databases.retrieve(db_uuid))
            self.cache[db.id] = db
        _logger.info(f'Retrieved database `{db.title}`.')
        return db

    def get_or_create_db(self, parent: Page, schema: type[Schema]) -> Database:
        """Get or create the database."""
        dbs = SList(db for db in self.search_db(schema._db_title) if db.parent == parent)
        if len(dbs) == 0:
            db = self.create_db(parent, schema=schema)
            while not [db for db in self.search_db(schema._db_title) if db.parent == parent]:
                _logger.info(f'Waiting for database `{db.title}` to be fully created.')
                time.sleep(1)
            return db
        else:
            db = dbs.item()
            db.schema = schema
            return db

    def search_page(self, title: str | None = None, *, exact: bool = True, reverse: bool = False) -> SList[Page]:
        """Search a page by name. Deleted pages, i.e. in trash, are not included in the search.

        Args:
            title: title of the page, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
        """
        _logger.info(f'Searching for page with title `{title}`.')
        query = cast(obj_query.SearchQueryBuilder[obj_blocks.Page], self.api.search(title).filter(page_only=True))
        if reverse:
            query.sort(ascending=True)
        pages = [
            cast(Page, self.cache.setdefault(raise_unset(page_obj.id), Page.wrap_obj_ref(page_obj)))
            for page_obj in query.execute()
        ]
        if exact and title is not None:
            pages = [page for page in pages if page.title == title]
        return SList(pages)

    def get_page(self, page_ref: UUID | str, *, use_cache: bool = True) -> Page:
        """Retrieve a page by uuid."""
        page_uuid = get_uuid(page_ref)
        if use_cache and page_uuid in self.cache:
            _logger.info(f'Retrieving cached page with id `{page_uuid}`.')
            page = cast(Page, self.cache[page_uuid])
        else:
            _logger.info(f'Retrieving page with id `{page_uuid}`.')
            try:
                page = Page.wrap_obj_ref(self.api.pages.retrieve(page_uuid))
            except APIResponseError as e:
                msg = f'Page with id {page_uuid} not found!'
                _logger.warning(msg)
                raise UnknownPageError(msg) from e
            self.cache[page.id] = page
        _logger.info(f'Retrieved page `{page.title}`.')
        return page

    def create_page(
        self, parent: Page | Database, title: Text | str | None = None, blocks: Sequence[Block] | None = None
    ) -> Page:
        """Create a new page in a `parent` page or database with a given `title`.

        The `blocks` are optional and can be used to create a page with content right away.
        Note that some nested blocks may not be supported by the API and must be created separately, i.e.
        with an `append` call to a given block.
        """
        _logger.info(f'Creating page with title `{title}` in parent `{parent.title}`.')
        title_obj = title if title is None else Title(title).obj_ref
        # We don't use the `children` parameter as we would need to call `list` afterwards to get the children,
        # in order to initialize them, which would be another API call. So we append the blocks manually here.
        page = Page.wrap_obj_ref(self.api.pages.create(parent=parent.obj_ref, title=title_obj))
        self.cache[page.id] = page

        if blocks:
            blocks_iter = _chunk_blocks_for_api(page, blocks)
            _append_block_chunks(blocks_iter)

        return page

    def get_or_create_page(self, parent: Page | Database, title: str | None = None) -> Page:
        """Get an existing page or create a new one if it doesn't exist."""
        pages = SList(page for page in self.search_page(title) if page.parent == parent)
        if len(pages) == 0:
            page = self.create_page(parent, title=title)
            while not [page for page in self.search_page(title) if page.parent == parent]:
                _logger.info(f'Waiting for page `{page.title}` to be fully created.')
                time.sleep(1)
            return page
        else:
            return pages.item()

    def get_user(self, user_ref: UUID | str, *, use_cache: bool = True, raise_on_unknown: bool = True) -> User:
        """Get a user by uuid.

        In case the user is not found and `raise_on_unknown` is `False`, an `User` object is returned
        with the name `Unknown User`, where the property `is_unknown` is set to `True`.

        !!! warning

            Trying to retrieve yourself, i.e. the bot integration, only works if `use_cache` is true,
            since the low-level api, i.e. `api.users.retrieve()` does not work for the bot integration.
            Better use `whoami()` to get the bot integration user object.
        """
        user_uuid = get_uuid(user_ref)
        self.whoami()  # make sure cache is filled with the uuid of the bot integration

        if use_cache and user_uuid in self.cache:
            _logger.info(f'Retrieving cached user with id `{user_uuid}`.')
            user = cast(User, self.cache[user_uuid])
        else:
            _logger.info(f'Retrieving user with id `{user_uuid}`.')
            try:
                user_obj = self.api.users.retrieve(user_uuid)
            except APIResponseError as e:
                msg = f'User with id {user_uuid} not found!'
                _logger.warning(msg)
                if raise_on_unknown:
                    raise UnknownUserError(msg) from e
                user_obj = UnknownUserObj(id=user_uuid, object='user', type='unknown')
            user = User.wrap_obj_ref(user_obj)
            self.cache[user.id] = user

        _logger.info(f'Retrieved user `{user}`.')
        return user

    def search_user(self, name: str) -> SList[User]:
        """Search a user by name."""
        _logger.info(f'Searching for user with name `{name}`.')
        return SList(user for user in self.all_users() if user.name == name)

    def all_users(self) -> list[User]:
        """Retrieve all users of this workspace."""
        _logger.info('Retrieving all users.')
        return [cast(User, self.cache.setdefault(user.id, User.wrap_obj_ref(user))) for user in self.api.users.list()]

    def whoami(self) -> Bot:
        """Return the integration as bot object."""
        _logger.info('Retrieving information about this integration bot.')
        if self._own_bot_id is None:
            user = self.api.users.me()
            self._own_bot_id = user.id
            return cast(Bot, self.cache.setdefault(user.id, Bot.wrap_obj_ref(user)))
        else:
            return cast(Bot, self.cache[self._own_bot_id])

    def upload(self, file: BinaryIO, *, file_name: str | None = None, mime_type: str | None = None) -> UploadedFile:
        """Upload a file to Notion."""
        file_name = file_name if file_name is not None else os.path.basename(getattr(file, 'name', 'unknown_file'))
        mime_type = mime_type if mime_type is not None else get_mime_type(file)
        file_size = get_file_size(file)
        if mime_type == 'application/octet-stream':
            _logger.warning(f'File `{file_name}` has unknown MIME type, falling back to text/plain.')
            mime_type = 'text/plain'  # Notion does not support application/octet-stream
        mode = FileUploadMode.SINGLE_PART if file_size <= MAX_FILE_SIZE else FileUploadMode.MULTI_PART
        _logger.info(
            f'Uploading file `{file_name}` of size {file_size} bytes with MIME type `{mime_type}` in mode `{mode}`.'
        )
        n_parts = -(-file_size // MAX_FILE_SIZE)  # ceiling division
        file_upload_obj = self.api.uploads.create(
            name=file_name, n_parts=None if n_parts == 1 else n_parts, mode=mode, content_type=mime_type
        )

        if mode == FileUploadMode.SINGLE_PART:
            self.api.uploads.send(file_upload=file_upload_obj, file=file)
        else:
            for part in range(1, n_parts + 1):
                _logger.info(f'Uploading part {part}/{n_parts} of file `{file_name}`.')
                chunk = file.read(MAX_FILE_SIZE)
                self.api.uploads.send(file_upload=file_upload_obj, part=part, file=io.BytesIO(chunk))
            self.api.uploads.complete(file_upload=file_upload_obj)

        return UploadedFile.from_file_upload(file_upload_obj)

    def import_url(self, url: str, file_name: str, *, block: bool = True) -> UploadedFile:
        """Import a file from a URL."""
        _logger.info(f'Importing file from URL `{url}`.')
        file_upload_obj = self.api.uploads.create(name=file_name, mode=FileUploadMode.EXTERNAL_URL, external_url=url)
        file_upload = UploadedFile.from_file_upload(file_upload_obj)
        if block:
            file_upload.wait_until_uploaded()
        return file_upload

    def list_uploads(self, filter: FileUploadStatus | None = None) -> list[UploadedFile]:  # noqa: A002
        """List all uploaded files and optionally filter by status."""
        _logger.info('Listing all uploaded files.')
        return [UploadedFile.from_file_upload(upload) for upload in self.api.uploads.list(status=filter)]



"""Google adapters."""
# ToDo: Extract some more functionality from the Google Task adapter into this level
# like GObject and the client code if it is the same over all Google adapters.

from ultimate_notion.adapters.google.tasks import GTasksClient, SyncGTasks

__all__ = ['GTasksClient', 'SyncGTasks']



import os
import string
from importlib.resources import read_text
from types import ModuleType


def get_template(name: str, relative_to: str | ModuleType = __name__) -> string.Template:
    """Retrieve the template by name"""
    if isinstance(relative_to, ModuleType):
        relative_to = relative_to.__name__

    data = read_text(relative_to, name, encoding='utf-8')
    # we assure that line endings are converted to '\n' for all OS
    content = data.replace(os.linesep, '\n')
    return string.Template(content)


def page_html(content: str, *, title: str = '') -> str:
    """Wrap the content of the page as proper HTML for displaying."""
    template = get_template('page.html')
    return template.safe_substitute(content=content, title=title)
